import { PathLike } from "fs";
import fs from "fs/promises";
import path from "path";
import sharp from "sharp";

const SUPPORTED_FORMATS = [
  ".png",
  ".jpg",
  ".jpeg",
  ".gif",
  ".bmp",
  ".webp",
  ".tiff",
];

export enum Orientation {
  TOP = 1,
  LEFT = 2,
  BOTTOM = 3,
  RIGHT = 4,
}

export interface SpeechBubbleOptions {
  mirror?: boolean;
  orientation?: Orientation;
  speechBubblePath?: string;
}

export class ImageSpeechBubbleTransformer {
  private defaultSpeechBubblePath: string;

  constructor(assetsDir: string = path.join(__dirname, "assets")) {
    this.defaultSpeechBubblePath = path.join(assetsDir, "speech_bubble.png");
  }

  /**
   * Check if the file format is supported
   */
  private checkFileFormat(filePath: string): void {
    const ext = path.extname(filePath).toLowerCase();
    if (!SUPPORTED_FORMATS.includes(ext)) {
      throw new Error(
        `Unsupported file extension ${ext}. Only ${SUPPORTED_FORMATS.join(", ")} are supported.`
      );
    }
  }

  /**
   * Transform image with optional mirroring and rotation
   */
  private async transformImage(
    inputBuffer: Buffer,
    options: { mirror?: boolean; orientation?: Orientation } = {}
  ): Promise<Buffer> {
    let pipeline = sharp(inputBuffer);

    // Mirror (flip horizontally)
    if (options.mirror) {
      pipeline = pipeline.flop();
    }

    // Rotate based on orientation
    switch (options.orientation) {
      case Orientation.LEFT:
        pipeline = pipeline.rotate(270);
        break;
      case Orientation.BOTTOM:
        pipeline = pipeline.rotate(180);
        break;
      case Orientation.RIGHT:
        pipeline = pipeline.rotate(90);
        break;
      default:
        // No rotation for TOP or default
        break;
    }

    return pipeline.toBuffer();
  }

  /**
   * Process image with speech bubble
   */
  async processSpeechBubble(
    inputBuffer: Buffer,
    options: SpeechBubbleOptions = {}
  ): Promise<Buffer> {
    if (!inputBuffer) {
      throw new Error("Input buffer is required");
    }

    const speechBubblePath =
      options.speechBubblePath || this.defaultSpeechBubblePath;

    this.checkFileFormat(speechBubblePath);

    const speechBubbleBuffer = await sharp(speechBubblePath).toBuffer();

    const processedImage = await sharp(inputBuffer)
      .ensureAlpha()
      .metadata()
      .then(async (metadata) => {
        // match image size
        const resizedSpeechBubble = await sharp(speechBubbleBuffer)
          .resize(metadata.width, metadata.height)
          .toBuffer();

        const transformedSpeechBubble = await this.transformImage(
          resizedSpeechBubble,
          {
            mirror: options.mirror,
            orientation: options.orientation,
          }
        );

        // subtract speech bubble
        return sharp(inputBuffer)
          .composite([
            {
              input: transformedSpeechBubble,
              blend: "difference",
            },
          ])
          .toBuffer();
      });

    return processedImage;
  }

  /**
   * Convenience method to process and save image
   */
  async processAndSave(
    inputBuffer: Buffer,
    outputPath: PathLike,
    options: SpeechBubbleOptions = {}
  ): Promise<void> {
    this.checkFileFormat(outputPath.toString());

    // Process image
    const processedBuffer = await this.processSpeechBubble(
      inputBuffer,
      options
    );

    const outputDir = path.dirname(outputPath.toString());
    await fs.mkdir(outputDir, { recursive: true });

    await sharp(processedBuffer).toFile(outputPath.toString());
  }
}

export function createSpeechBubbleTransformer(assetsDir?: string) {
  return new ImageSpeechBubbleTransformer(assetsDir);
}
